#!/usr/bin/env python
# coding: utf-8

import pandas as pd
from pprint import pformat
import scipy.sparse
import anndata as ad

from ._classify import classify, assert_demux
from ._defaults import DEFAULTS, DESCRIPTIONS
from ._logging import get_logger
from ._meta import add_meta
from ._utils import get_layer

def heuristics(
    adata_hto: ad.AnnData,
    inplace: bool = DEFAULTS["inplace"],
    verbose: int = DEFAULTS["verbose"],
    key_normalise: str = DEFAULTS["add_key_normalise"],
    key_denoise: str = DEFAULTS["add_key_denoise"],
    add_key_hashid: str = DEFAULTS["add_key_hashid"],
    add_key_doublet: str = DEFAULTS["add_key_doublet"],
):
    raise NotImplementedError("Heuristics function is not implemented yet.")


import click
from hto import dnd
from hto._defaults import OPTIONS

# create function
@click.command(
    name="dnd",
    help="Perform normalisation and demultiplexing of HTO data.",
    no_args_is_help=True,
)
def cli(**kwargs):
    dnd(
        _as_cli=True,
        **kwargs
    )

# add options
for key, option in OPTIONS.items():
    # skip anonymous options
    if key[0] == "_":
        continue
    cli = option(cli)

if __name__ == "__main__":
    cli()


import os
import numpy as np
import anndata as ad
import pytest
from pandas.api.types import is_float_dtype, is_integer_dtype

from hto import normalise
from hto._exceptions import AnnDataFormatError, UserInputError

@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_normalise(mock_hto_data):
    """
    Test if normalisation works.
    """
    # Get mock data
    adata_filtered = mock_hto_data['filtered']
    adata_raw = mock_hto_data['raw']

    # Run normalisation
    adata_norm = normalise(
        adata_hto=adata_filtered,
        adata_hto_raw=adata_raw,
        add_key_normalise=None,
        inplace=False,
        background_version="v2",
    )

    # Test layers, datatypes, shapes and metadata
    assert adata_norm.shape == adata_filtered.shape
    assert len(adata_norm.layers) == 0
    assert isinstance(adata_norm.X, np.ndarray)
    assert np.any(adata_norm.X < 0)
    assert is_float_dtype(adata_norm.X)
    assert is_integer_dtype(adata_filtered.X)
    assert is_integer_dtype(adata_raw.X)
    assert "dnd" in adata_norm.uns
    assert "normalise" in adata_norm.uns["dnd"]
    mu = adata_norm.uns["dnd"]["normalise"]["mu_empty"]
    assert np.all(mu < 5) and np.all(mu > 1)
    assert adata_norm.uns["dnd"]["normalise"]["params"]["pseudocount"] == 10

    # Test inplace
    adata_inplace = adata_filtered.copy()
    normalise(
        adata_hto=adata_inplace,
        adata_hto_raw=adata_raw,
        add_key_normalise=None,
        pseudocount=20,
        inplace=True,
        background_version="v2",
    )
    assert is_float_dtype(adata_inplace.X)
    assert adata_inplace.uns["dnd"]["normalise"]["params"]["pseudocount"] == 20

    # Test layers
    adata_norm = normalise(
        adata_hto=adata_filtered,
        adata_hto_raw=adata_raw,
        add_key_normalise="norm",
        inplace=False,
        background_version="v2",
    )
    assert "norm" in adata_norm.layers


@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_faulty_data(mock_hto_data):
    """
    Test if normalisation works.
    """
    # Get mock data
    adata_filtered = mock_hto_data['filtered']
    adata_raw = mock_hto_data['raw']

    # Fail - full overlap
    with pytest.raises(UserInputError):
        adata_norm = normalise(
            adata_hto=adata_filtered,
            adata_hto_raw=adata_filtered,
            inplace=False,
            _run_assert=False,
            background_version="v2",
        )

    # Fail - no overlap
    with pytest.raises(UserInputError):
        obs_names = adata_raw.obs_names
        obs_names = obs_names[~obs_names.isin(adata_filtered.obs_names)]
        adata_norm = normalise(
            adata_hto=adata_filtered,
            adata_hto_raw=adata_raw[obs_names],
            inplace=False,
            background_version="v2",
        )



from . import tl
from . import data
from . import pl
from .normalise import normalise, normalise_debug
from .denoise import denoise
from .demux import demux
from .dnd import dnd

__all__ = ["normalise", "denoise", "demux", "dnd", "tl", "data", "pl"]


import numpy as np

class UserInputError(Exception):
    pass

class AnnDataFormatError(Exception):
    def __init__(self, key, *args, **kwargs):
        message = self.get_message(key, *args, **kwargs)
        super().__init__(message)

    def get_message(self, key, *args, **kwargs):
        _func = {
            "adata_raw_missing_cells": self._adata_raw_missing_cells,
            "adata_no_overlapping_names": self._adata_filtered_too_few_cells,
            "adata_not_float": self._adata_not_float,
        }.get(key, self._default)
        return _func(*args, **kwargs)

    def _default(self, *args, **kwargs):
        args = ", ".join(args)
        kwargs = ", ".join([f"{k}={v}" for k, v in kwargs.items()])
        return f"Unknown error occurred. Args: {args}, kwargs: {kwargs}"

    def _adata_raw_missing_cells(self, empty_barcodes):
        return f"Too few empty droplets detected: '{len(empty_barcodes)}'. " \
               f"Make sure that 'adata_raw' contains cells that are not present in 'adata_filtered'."

    def _adata_filtered_too_few_cells(self, n_cells):
        return f"Too few cells in 'adata_filtered': '{n_cells}'. "

    def _adata_not_float(self, x):
        msg = f"Input matrix must contain float values, got {x.dtype} ({', '.join(np.unique(x).astype(str)[:10])})."
        return msg



import numpy as np
from scipy.stats import skewnorm
from scipy.optimize import curve_fit, Bounds
from numpy.typing import ArrayLike
from skimage.filters import threshold_otsu
import matplotlib.pyplot as plt
import seaborn as sns

from ._logging import get_logger

def _estimate_initial_params(
    data: ArrayLike,
    n_probes: int,
    threshold: float = None,
) -> tuple:
    """
    Estimate initial parameters for a skew-normal mixture model.
    """

    # setup
    if threshold is None:
        threshold = threshold_otsu(data)

    # inits
    lower_w = (n_probes - 1) / n_probes
    lower_skew = 1
    lower_loc = np.mean(data[data < threshold])
    lower_sd = np.std(data[data < threshold])

    upper_w = 1 / n_probes
    upper_skew = -1
    upper_loc = np.mean(data[data > threshold])
    upper_sd = np.std(data[data > threshold])

    init = [
        lower_w, lower_skew, lower_loc, lower_sd,
        upper_w, upper_skew, upper_loc, upper_sd,
    ]

    return init


def skewnorm_mixture_pdf(
    x: ArrayLike,
    s1: float, a1: float, loc1: float, scale1: float,
    s2: float, a2: float, loc2: float, scale2: float,
) -> ArrayLike:
    """
    Get pdf of two skew-normal distributions.

    Parameters
    ----------
    x : ArrayLike
        Input data points.
    s1, s2 : float
        Scale factors for the first and second skew-normal distributions.
    a1, a2 : float
        Skew parameters for the first and second skew-normal distributions.
    loc1, loc2 : float
        Location parameters (means) for the first and second distributions.
    scale1, scale2 : float
        Scale parameters (standard deviations) for the first and second
        distributions.

    Returns
    -------
    ArrayLike
        Sum of densities from the two skew-normal distributions.
    """
    density_1 = s1 * skewnorm.pdf(x, a=a1, loc=loc1, scale=scale1)
    density_2 = s2 * skewnorm.pdf(x, a=a2, loc=loc2, scale=scale2)
    return density_1,  density_2

def _skewnorm_sum(*args, **kwargs):
    d1, d2 = skewnorm_mixture_pdf(*args, **kwargs)
    return d1 + d2

def _assign_skewnorm_proba(
    x: ArrayLike,
    s1: float, a1: float, loc1: float, scale1: float,
    s2: float, a2: float, loc2: float, scale2: float,
):
    """Return the probability of each data point coming from the second skew-normal distribution."""
    density_1, density_2 = skewnorm_mixture_pdf(
        x, s1, a1, loc1, scale1, s2, a2, loc2, scale2
    )
    # hard cutoff at centers
    density_sum = density_2 / (density_1 + density_2 + 1e-4)
    density_sum[x < loc1] = 0
    density_sum[x > loc2] = 1
    return density_sum

def skewnorm_mixture_model(
    data: ArrayLike,
    p0: list = None,
    nbins: int = 100,
    p_cutoff: float = 0.95,
    n_probes: int = None,
    verbose: bool = False,
) -> ArrayLike:
    """
    Label data points above a cutoff using a mixture of skew-normal
    distributions.

    Parameters
    ----------
    data : ArrayLike
        Input data to label.
    qtl_cutoff : float, optional
        Quantile cutoff to classify positive labels. Default is 0.5.
    nbins : int, optional
        Number of bins for the histogram used in fitting. Default is 100.
    return_params : bool, optional
        Whether to return fitted parameters alongside labels. Default is False.

    Returns
    -------
    ArrayLike
        Boolean array indicating whether each data point is labeled positive.
        If `return_params` is True, also returns the fitted parameters.
    """

    # Get logger
    logger = get_logger("skewnorm_mixture_model", level=verbose)

    # Generate histogram and normalize to match PDF
    counts, edges = np.histogram(data, bins=nbins)
    centers = (edges[:-1] + edges[1:]) / 2
    counts = counts.astype(float)
    #counts = np.log10(counts + 1)  # Log transformation to stabilize counts
    bin_width = edges[1] - edges[0]
    counts /= counts.sum() * bin_width

    # Initial parameters
    if p0 is None:
        assert n_probes is not None, "n_probes must be provided if p0 is not."
        p0 = _estimate_initial_params(data, n_probes=n_probes)

    # Bounds
    bounds = np.array([
        [0, 32],         # s1, scale of first skewnorm
        [0, 5],        # a1, skew of first skewnorm
        [-5, 5],  # loc1, mean of first skewnorm
        [0.1, 5],        # scale1, std of first skewnorm
        [0, 32],         # s2, scale of second skewnorm
        [-1, 0],       # a2, skew of second skewnorm
        [0, 8],  # loc2, mean of second skewnorm
        [0.1, 5],        # scale2, std of second skewnorm
    ]).T
    bounds = Bounds(*bounds)

    # Fit parameters
    try:
        params, _ = curve_fit(
            _skewnorm_sum,
            centers,
            counts,
            p0=p0,
            bounds=bounds,
            method='dogbox',
        )
    except RuntimeError:
        logger.warning("Fitting failed. Data quality low. Using Otsu's method, but results may be unreliable.")
        params = p0

    # Calculate probabilities
    probs = _assign_skewnorm_proba(data, *params)

    # Get thresholds where the probability is above a certain cutoff
    if len(data[probs > p_cutoff]) == 0:
        threshold = data.max()
    else:
        threshold = np.min(data[probs > p_cutoff])

    return probs, threshold, params

def plot_skewnorm(
    data,
    params,
    nbins=100,
    ax=None,
    **kwargs
):
    """
    Plot both skew-normal distributions and the data as a histogram.
    """

    # setup
    if ax is None:
        fig, ax = plt.subplots()
    bins = np.linspace(data.min(), data.max(), nbins)

    # plot densities
    #sns.kdeplot(data, fill=True, alpha=.2, color=kwargs.pop("color", "orange"), ax=ax)
    # histogram
    counts = np.histogram(data, bins=nbins)[0]
    bin_width = bins[1] - bins[0]
    counts = counts.astype(float)
    counts /= counts.sum() * bin_width
    ax.bar(bins, counts, width=bin_width, alpha=0.5, color="grey")

    # plot skewnorms
    d_low, d_high = skewnorm_mixture_pdf(bins, *params)
    cell_proba = _assign_skewnorm_proba(bins, *params)
    ax.plot(bins, d_low, label="Low", color="red")
    ax.plot(bins, d_high, label="High", color="blue")
    ax.plot(bins, cell_proba, label="Probability Signal", color="grey")

    # plot data

    # legend
    ax.legend()

    return ax


import numpy as np
from sklearn.linear_model import LinearRegression
import scipy
import anndata as ad
from pandas.api.types import is_integer_dtype, is_float_dtype

from ._cluster_background import assert_background, estimate_background
from ._remove_batch_effect import remove_batch_effect

from ._meta import add_meta
from ._logging import get_logger
from ._defaults import DEFAULTS, DESCRIPTIONS
from ._exceptions import UserInputError
from ._utils import get_layer

def denoise(
    adata_hto: ad.AnnData,
    use_layer: str = DEFAULTS["use_layer"],
    background_method: str = DEFAULTS["background_method"],
    add_key_denoise: str = DEFAULTS["add_key_denoise"],
    covariates: np.ndarray = DEFAULTS["covariates"],
    denoise_version: str = DEFAULTS["denoise_version"],
    design: np.ndarray = DEFAULTS["design"],
    inplace: bool = DEFAULTS["inplace"],
    verbose: int = DEFAULTS["verbose"],
    kwargs_denoise: dict = DEFAULTS["kwargs_denoise"],
    **kwargs,
):
    f"""Remove technical noise by regressing out cell-level background expression.

    This function aims to remove technical noise from normalized protein expression data:
    1. Build Background Data: Estimate the technical noise per cell using kmeans or GMM.
    2. Remove Batch Effect: Fit a linear model to the normalized matrix and subtract the
        technical noise from the original matrix.

    Args:
        adata_hto (ad.AnnData): {DESCRIPTIONS["adata_hto"]}
        use_layer (str, optional): {DESCRIPTIONS["use_layer"]}
        background_method (str, optional): {DESCRIPTIONS["background_method"]}
        covariates (np.ndarray, optional): {DESCRIPTIONS["covariates"]}
        design (np.ndarray, optional): {DESCRIPTIONS["design"]}
        add_key_denoise (str, optional): {DESCRIPTIONS["add_key_denoise"]}
        inplace (bool, optional): {DESCRIPTIONS["inplace"]}
        verbose (int, optional): {DESCRIPTIONS["verbose"]}

    Returns:
        ad.AnnData: An updated AnnData object with denoised protein expression data.
    """
    assert_background(
        method=background_method,
        **kwargs
    )

    # logger
    logger = get_logger("denoise", level=verbose)
    logger.log_parameters(locals())
    logger.debug("Starting denoising...")

    # setup data
    adata_hto, x = get_layer(
        adata=adata_hto,
        use_layer=use_layer,
        numpy=True,
        float=True,
        inplace=inplace,
    )

    # 1. Build Background Data
    if covariates is None:
        logger.debug(f"Build background data using '{background_method}' method")
        covariates, meta_background = estimate_background(
            matrix=x,
            method=background_method,
            **kwargs,
        )
    else:
        meta_background = {}

    # 2. Remove Batch Effect
    logger.debug("Removing technical noise")
    norm_adt, meta_batch_model = remove_batch_effect(
        denoise_version,
        x=x,
        covariates=covariates,
        design=design,
        verbose=verbose,
        kwargs_denoise=kwargs_denoise,
    )

    # Store meta information (don't use 'debug' key)
    logger.debug("Technical noise removal completed.")
    #meta_background = {k: v for k, v in meta_background.items() if k != "debug"}
    adata_hto = add_meta(
        adata_hto,
        step="denoise",
        params={
            "background_method": background_method,
        },
        covariates=covariates,
        batch_model=meta_batch_model,
        meta_background=meta_background,
    )

    # Finish
    if add_key_denoise is not None:
        adata_hto.layers[add_key_denoise] = norm_adt
        logger.info(f"Denoised matrix stored in adata.layers['{add_key_denoise}']")
    else:
        adata_hto.X = norm_adt
        logger.info("Denoised matrix stored in adata.X")


    return adata_hto



import os
import numpy as np
import anndata as ad
import pytest

from hto import dnd
from hto._exceptions import AnnDataFormatError

@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_dnd(mock_hto_data):
    """
    Test the full pipeline: normalisation, denoising, and demultiplexing.
    """

    # Get mock data
    adata_filtered = mock_hto_data['filtered']
    adata_raw = mock_hto_data['raw']

    # Not in place
    adata_result = dnd(
        adata_hto=adata_filtered,
        adata_hto_raw=adata_raw,
        background_method="kmeans-fast",
        add_key_normalise="normalised",
        add_key_denoise="denoised",
        background_version="v2",
        pseudocount=20,
        inplace=False,
    )

    # In place
    adata_filtered_copy = adata_filtered.copy()
    dnd(
        adata_hto=adata_filtered_copy,
        adata_hto_raw=adata_raw,
        background_method="kmeans-fast",
        add_key_normalise="normalised",
        add_key_denoise="denoised",
        background_version="v2",
        pseudocount=20,
        inplace=True,
    )

    # assert both
    def assert_results(adata):
        # Verify DSB output
        assert "normalised" in adata.layers, "Normalised layer not found in output"
        assert "denoised" in adata.layers, "Denoised layer not found in output"

        # Verify demultiplexing output
        assert isinstance(adata, ad.AnnData), "Demultiplexing result is not an AnnData object"
        assert 'hash_id' in adata.obs.columns, "'hash_id' column not found in demultiplexing result"
        assert 'doublet_info' in adata.obs.columns, "'doublet_info' column not found in demultiplexing result"
        assert len(adata.obs.columns) == 5, f"Expected columns cell_id', 'cell_type', 'hto_classification', 'hash_id', 'doublet_info'. Unexpected columns in demultiplexing result: {adata.obs.columns}"
        assert 'metrics' in adata.uns["dnd"]["demux"], "Metrics not found in demultiplexing result"
        assert 'pseudocount' in adata.uns["dnd"]["normalise"]["params"], "Pseudo count not found in normalisation result"
        assert adata.uns["dnd"]["normalise"]["params"]["pseudocount"] == 20, "Pseudo count is not 20"

        # Check classifications
        hash_ids = adata.var_names
        classifications = adata.obs['hash_id'].value_counts()
        assert len(classifications) > 1, "No classifications found"
        assert all(classifications.index.isin(['negative', 'doublet'] + list(hash_ids))), "Invalid classifications found"

        # Check if metrics exist for each HTO
        metrics = adata.uns["dnd"]["demux"]["metrics"]
        assert all(i in metrics for i in hash_ids), "Metrics missing for some HTOs"

        # Check overall accuracy
        true_labels = adata.obs['hto_classification']
        predicted_labels = adata.obs['hash_id']
        overall_accuracy = np.mean(predicted_labels == true_labels)
        assert overall_accuracy > 0.8, f"Overall accuracy is only {overall_accuracy:.2f}"

    assert_results(adata_result)
    assert_results(adata_filtered_copy)


@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_no_background(mock_hto_data):

    # Get mock data
    adata_filtered = mock_hto_data['filtered']
    adata_raw = adata_filtered.copy()

    # Test with no background
    with pytest.raises(AnnDataFormatError):
        adata = dnd(
            adata_hto=adata_filtered,
            adata_hto_raw=adata_filtered.copy(),
            adata_gex=adata_filtered.copy(),
        )

    # Test with too few cells
    with pytest.raises(AnnDataFormatError):
        adata = dnd(
            adata_hto=adata_filtered[:2],
            adata_hto_raw=adata_raw,
            adata_gex=adata_filtered,
        )


from pprint import pformat
import numpy as np
import anndata as ad
from typing import Union
from ._logging import get_logger
from ._meta import init_meta, add_meta
from ._exceptions import AnnDataFormatError, UserInputError
from ._defaults import DEFAULTS, DESCRIPTIONS
from ._utils import get_layer
from .tl import build_background


_normalise_version_meta = {
    "v1": {
        "required": ["adata_hto"],
        "optional": ["adata_hto_raw", "adata_gex", "adata_background"],
        "description": "Background aware HTO normalisation.",
    }
}

def assert_normalisation(df, logger, max_spread=1.5, qs=[.1, .99]):
    """Assert that normalised values are within a certain spread."""
    spans = []
    # get spread of each column
    for c in df.columns:
        q = df[c].quantile(qs).values
        diff = float(q[1] - q[0])
        spans.append(diff)
    ratio = max(spans) / min(spans)
    if ratio > max_spread:
        logger.warning(f"Spread of normalised values is too high: {ratio:.2f}. This may cause issues when estimate cell-by-cell covariates.")

def normalise(
    adata_hto: ad.AnnData,
    adata_hto_raw: ad.AnnData = None,
    adata_gex: ad.AnnData = None,
    adata_background: ad.AnnData = None,
    pseudocount: int = DEFAULTS["pseudocount"],
    background_version: str = DEFAULTS["background_version"],
    add_key_normalise: str = DEFAULTS["add_key_normalise"],
    use_layer: str = DEFAULTS["use_layer"],
    inplace: bool = DEFAULTS["inplace"],
    verbose: int = DEFAULTS["verbose"],
    **kwargs_background,
) -> ad.AnnData:
    f"""Background aware HTO normalisation.

    This function implements an adapted version of the DSB algorithm (see citation), which normalizes protein
    expression data using empty droplets as background reference and optionally performs
    technical noise removal.

    Args:
        adata_hto (AnnData): {DESCRIPTIONS["adata_hto"]}
        adata_hto_raw (AnnData): {DESCRIPTIONS["adata_hto_raw"]}
        adata_gex (AnnData, optional): {DESCRIPTIONS["adata_gex"]}
        adata_background (AnnData, optional): {DESCRIPTIONS["adata_background"]}
        pseudocount (int, optional): {DESCRIPTIONS["pseudocount"]}
        background_method (str, optional): {DESCRIPTIONS["background_method"]}
        add_key_normalise (str, optional): {DESCRIPTIONS["add_key_normalise"]}
        use_layer (str, optional): {DESCRIPTIONS["use_layer"]}
        inplace (bool, optional): {DESCRIPTIONS["inplace"]}
        verbose (int, optional): {DESCRIPTIONS["verbose"]}

    Returns:
        ad.AnnData: AnnData object with normalized protein expression data, either
        in the original matrix or in a specified layer.

    Citation:
    Mulè, M.P., Martins, A.J. & Tsang, J.S. Normalizing and denoising protein expression data from droplet-based single cell profiling. Nat Commun 13, 2099 (2022). https://doi.org/10.1038/s41467-022-29356-8
    """
    # Get logger
    logger = get_logger("normalise", level=verbose)
    logger.log_parameters(locals())
    logger.debug("Starting normalization...")

    # Subset to same columns in hto and hto_raw
    if not all(adata_hto.var_names.isin(adata_hto_raw.var_names)):
        raise UserInputError(f"Columns don't match. 'adata_hto' has columns that are not in 'adata_hto_raw'. Make sure that all var_names in 'adata_hto' are also in 'adata_hto_raw'.")
    adata_hto_raw = adata_hto_raw[:, adata_hto.var_names].copy()
    
    # Get background if not provided
    adata_background = build_background(
        background_version=background_version,
        adata_hto=adata_hto,
        adata_hto_raw=adata_hto_raw,
        adata_gex=adata_gex,
        adata_background=adata_background,
        verbose=verbose,
        **kwargs_background,
    )

    # Setup
    adata_hto, adt = get_layer(
        adata_hto,
        use_layer=use_layer,
        integer=True,
        numpy=True,
        inplace=inplace
    )

    adata_background, adtu = get_layer(
        adata_background,
        use_layer=use_layer,
        integer=True,
        numpy=True,
        inplace=inplace
    )

    # Init metadata
    adata_hto = init_meta(adata_hto)

    # Subsets
    barcodes_filtered = set(adata_hto.obs_names)
    barcodes_background = set(adata_background.obs_names)
    barcodes_background_only = barcodes_background.difference(barcodes_filtered)
    overlap_barcode = list(barcodes_filtered & barcodes_background)  # check that naming is consistent

    n_filtered = len(barcodes_filtered)
    n_background = len(barcodes_background)
    pct_background = n_filtered / n_background * 100

    logger.debug(f"Filtered adata: {n_filtered / 1000:.1f}K cells | Background adata: {n_background / 1000:.1f}K cells")
    logger.debug(f"Background cells: {n_background / 1000:f}K cells | Overlapping cells: {len(overlap_barcode) / 1000:f}K cells")
    if pct_background < 10:
        logger.warning(f"Only few barcodes are used for normalization: {n_background / 1000:.1f}K ({pct_background:.1f}%)")

    # Identify barcodes that are in adata_raw but not in adata_filtered
    if len(barcodes_background_only) < 5:
        raise AnnDataFormatError("adata_raw_missing_cells", barcodes_background)
    if len(overlap_barcode) < 5:
        raise AnnDataFormatError("adata_no_overlapping_names", len(barcodes_filtered))

    # Log transform both matrices
    adt_log = np.log(adt + pseudocount)
    adtu_log = np.log(adtu + pseudocount)

    # Calculate mean and sd of log-transformed empty droplets for each protein
    mu_empty = np.array(np.mean(adtu_log, axis=0)).flatten()
    sd_empty = np.array(np.std(adtu_log, axis=0)).flatten()

    # Normalize the cell protein matrix
    normalized_matrix = (adt_log - mu_empty) / sd_empty

    # Checkpoint
    if add_key_normalise is not None:
        adata_hto.layers[add_key_normalise] = normalized_matrix
        logger.info(f"Normalized matrix stored in adata.layers['{add_key_normalise}']")
    else:
        adata_hto.X = normalized_matrix
        logger.info("Normalization completed and stored in adata.X")

    # Assert
    assert_normalisation(adata_hto.to_df(add_key_normalise), logger)

    # Log metadata
    logger.debug(pformat(adata_hto.uns["dnd"]))

    # Store meta information
    adata_hto = add_meta(
        adata_hto,
        step="normalise",
        params={
            "pseudocount": pseudocount,
            "background": adata_background.obs_names.values,
        },
        mu_empty=mu_empty,
        sd_empty=sd_empty,
    )

    return adata_hto



def normalise_debug(
    adata_hto: ad.AnnData,
    background_quantile: float = DEFAULTS["background_quantile"],
    use_layer: str = DEFAULTS["use_layer"],
    verbose: int = DEFAULTS["verbose"],
):
    f"""
    If no background data or GEX data is available, align the quantiles of the
    filtered HTO data from 0 to 1. This is clearly not recommended, but is provided as a last
    resort option.

    Args:
        adata_hto (AnnData): {DESCRIPTIONS["adata_hto"]}
        background_quantile (float, tuple, optional): {DESCRIPTIONS["background_quantile"]}
        use_layer (str, optional): {DESCRIPTIONS["use_layer"]}
        verbose (int, optional): {DESCRIPTIONS["verbose"]}
    """
    logger = get_logger("utils", level=verbose)

    # estimate empty
    df = adata_hto.to_df(use_layer)
    adt_log = np.log1p(df).values

    mu_empty = []
    sd_empty = []
    for i in range(adt_log.shape[1]):
        q = np.quantile(adt_log[:, i], background_quantile)
        mu_temp = np.mean(adt_log[adt_log[:, i] < q, i])
        sd_temp = np.std(adt_log[adt_log[:, i] < q, i])
        mu_empty.append(mu_temp)
        sd_empty.append(sd_temp)

    # normalise
    normalized_matrix = (adt_log - mu_empty) / sd_empty
    adata_hto.layers["normalised"] = normalized_matrix
    return adata_hto



import numpy as np
import pytest
from pandas.api.types import is_float_dtype

from hto import normalise, denoise
from hto._exceptions import AnnDataFormatError
from hto._cluster_background import SUPPORTED_BACKGROUND_METHODS
from hto._remove_batch_effect import SUPPORTED_DENOISE_VERSIONS

@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_denoise(mock_hto_data):
    """
    Test if technical denoiseing works.
    """
    # Get mock data
    adata_filtered = mock_hto_data['filtered']
    adata_raw = mock_hto_data['raw']

    # Run normalisation
    adata_norm = normalise(
        adata_hto=adata_filtered,
        adata_hto_raw=adata_raw,
        add_key_normalise="normalised",
        background_version="v2",
        inplace=False,
    )

    # Run denoising
    adata_denoised = denoise(
        adata_hto=adata_norm,
        background_method="kmeans-fast",
        use_layer="normalised",
        add_key_denoise="denoised",
        inplace=False,
    )

    # Verify that the normalised and denoised layer is added and has expected dimensions
    assert "normalised" in adata_denoised.layers
    assert "denoised" in adata_denoised.layers
    assert adata_denoised.layers["normalised"].shape == adata_filtered.X.shape
    assert adata_denoised.layers["denoised"].shape == adata_filtered.X.shape

    # Check that denoised values are different from non-denoised values
    assert not np.array_equal(adata_denoised.layers["denoised"], adata_denoised.layers["normalised"]), \
        "Denoised and non-denoised results should be different"

    # Check that the denoised values have lower variance
    denoised_variance = np.var(adata_denoised.layers["denoised"])
    non_denoised_variance = np.var(adata_denoised.layers["normalised"])
    assert denoised_variance < non_denoised_variance, \
        "Denoised data should have lower variance than non-denoised data"

    # Assert that covariate vector is stored in 'uns'
    assert "denoise" in adata_denoised.uns["dnd"]
    assert "covariates" in adata_denoised.uns["dnd"]["denoise"]
    assert len(adata_denoised.uns["dnd"]["denoise"]["covariates"]) == adata_filtered.shape[0]

    # Test inplace
    adata_inplace = adata_filtered.copy()
    normalise(
        adata_hto=adata_inplace,
        adata_hto_raw=adata_raw,
        background_version="v2",
        add_key_normalise=None,
        inplace=True,
    )
    assert is_float_dtype(adata_inplace.X)

    denoise(
        adata_hto=adata_inplace,
        background_method="kmeans-fast",
        add_key_denoise=None,
        inplace=True,
    )
    assert is_float_dtype(adata_inplace.X)


@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_skip_normalise(mock_hto_data):
    """
    Test if technical denoiseing works.
    """
    # Get mock data
    adata_filtered = mock_hto_data['filtered']

    # Run denoising
    with pytest.raises(AnnDataFormatError):
        adata_denoised = denoise(
            adata_hto=adata_filtered,
            background_method="kmeans-fast",
            inplace=False,
        )

@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_background_methods(mock_hto_data):
    """
    Test if technical denoiseing works.
    """
    # Get mock data
    adata_filtered = mock_hto_data['filtered']
    adata_raw = mock_hto_data['raw']

    # Run normalisation
    adata_norm = normalise(
        adata_hto=adata_filtered,
        adata_hto_raw=adata_raw,
        add_key_normalise="normalised",
        inplace=False,
        background_version="v2",
    )

    adata_benchmark = denoise(
        adata_hto=adata_norm,
        background_method="kmeans-fast",
        use_layer="normalised",
        add_key_denoise="denoised",
        inplace=False,
    )
    covariates_benchmark = adata_benchmark.uns["dnd"]["denoise"]["covariates"]

    # Run denoising
    for method in SUPPORTED_BACKGROUND_METHODS:
        for denoise_version in SUPPORTED_DENOISE_VERSIONS:
            adata_denoised = denoise(
                adata_hto=adata_norm,
                background_method=method,
                use_layer="normalised",
                add_key_denoise="denoised",
                denoise_version=denoise_version,
                inplace=False,
            )
            covariates_test = adata_denoised.uns["dnd"]["denoise"]["covariates"]

            assert np.corrcoef(covariates_benchmark, covariates_test)[0, 1] > 0.9, \
                "Covariates from kmeans and kmeans-fast should be highly correlated"



"""Module containing functions for clustering and demultiplexing HTO data."""

import numpy as np
from sklearn.mixture import GaussianMixture
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score, davies_bouldin_score
from skimage.filters import threshold_otsu
import scipy.stats
from scipy.optimize import root_scalar

from ._logging import get_logger


SUPPORTED_DEMUX_METHODS = ["kmeans", "gmm", "otsu", "gmm_demux"]

def assert_demux(method):
    if method not in SUPPORTED_DEMUX_METHODS:
        raise ValueError(f"Method '{method}' not supported. Choose from: {', '.join(SUPPORTED_DEMUX_METHODS)}")

def classify(
    data,
    demux_method="kmeans",
    verbose=1,
):
    """Perform clustering to identify positive populations for each HTO.

    Methods:
        - K-means: Silhouette score and Davies-Bouldin index
        - GMM: BIC and log-likelihood
        - Otsu: Inter-class variance and entropy
        - GMM demux: Uses external gmm_demux function

    Args:
        data (np.array): The input data used for clustering
        method (str): Clustering method to use. Either 'kmeans' or 'gmm'. Default is 'kmeans'.

    Returns:
        tuple: A tuple containing:
            - np.array: The cluster labels
            - int: The index of the positive cluster
            - dict: A dictionary containing various goodness of fit metrics
    """
    logger = get_logger("demux", level=verbose)
    if demux_method == "kmeans":
        logger.debug("Applying K-means to each column")
        return classify_by_column(_classify_kmeans_one, data, logger)
    elif demux_method == "gmm":
        logger.debug("Applying GMM to each column")
        return classify_by_column(_classify_gmm_one, data, logger)
    elif demux_method == "otsu":
        logger.debug("Applying Otsu's method to each column")
        return classify_by_column(_classify_otsu_one, data, logger)
    elif demux_method == "gmm_demux":
        logger.debug("Applying GMM demux to each column")
        return classify_gmm_demux(data, logger)
    else:
        raise ValueError(f"Method '{demux_method}' is not supported. Must be one of {SUPPORTED_DEMUX_METHODS}")


def classify_by_column(func, df_umi, logger):
    """Demultiplex HTO data by clustering each column. Reusable for different methods."""
    # init
    classifications = {}
    metrics = {}
    thresholds = {}

    for hto in df_umi.columns:
        # prepare data
        logger.debug(f"Demultiplexing HTO '{hto}'...")
        data = df_umi[hto].values.reshape(-1, 1)
        # apply function
        labels, threshold, hto_metrics = func(data, logger)
        # store results
        thresholds[hto] = float(threshold)
        metrics[hto] = hto_metrics
        classifications[hto] = labels
    
    return classifications, thresholds, metrics


def _classify_kmeans_one(series, logger):
    # predict clusters
    logger.debug("Clustering HTO data using K-means")
    model = KMeans(n_clusters=2, random_state=42, n_init=10)
    cluster = model.fit_predict(series)
    positive_cluster = np.argmax(model.cluster_centers_)
    labels = (cluster == positive_cluster).astype(int)

    # evaluate
    logger.debug("Evaluating K-means clustering")
    threshold = np.mean(model.cluster_centers_)
    silhouette = silhouette_score(series, labels)
    davies_bouldin = davies_bouldin_score(series, labels)
    metrics = {"silhouette_score": float(silhouette), "davies_bouldin_index": float(davies_bouldin)}
    return labels, threshold, metrics

def _classify_gmm_one(series, logger):
    # predict clusters
    logger.debug("Clustering HTO data using GMM")
    model = GaussianMixture(n_components=2, random_state=42)
    cluster = model.fit_predict(series)
    positive_cluster = np.argmax(model.means_)
    labels = (cluster == positive_cluster).astype(int)
    
    # extract GMM parameters - intersection of two Gaussians
    logger.debug("Evaluating GMM clustering")
    means, variances, weights = map(np.ravel, (model.means_, model.covariances_, model.weights_))
    sorted_indices = np.argsort(means)
    mu1, mu2 = means[sorted_indices]
    sigma1, sigma2 = np.sqrt(variances[sorted_indices])
    pi1, pi2 = weights[sorted_indices]
    def pdf_diff(x):
        return (pi1 / sigma1) * np.exp(-0.5 * ((x - mu1) / sigma1)**2) - (pi2 / sigma2) * np.exp(-0.5 * ((x - mu2) / sigma2)**2)
    threshold = root_scalar(pdf_diff, bracket=[mu1, mu2]).root
    
    # evaluate
    logger.debug("Evaluating GMM clustering")
    bic = model.bic(series)
    log_likelihood = model.score(series) * series.shape[0]
    metrics = {"bic": float(bic), "log_likelihood": float(log_likelihood)}

    return labels, threshold, metrics


def _classify_otsu_one(series, logger):
    # predict
    logger.debug("Thresholding HTO data using Otsu's method")
    series = series.flatten()
    threshold = threshold_otsu(series)
    labels = (series > threshold).astype(int)

    # evaluate
    logger.debug("Evaluating Otsu thresholding")
    background = series[labels == 0]
    signal = series[labels == 1]
    # Inter-class variance (which Otsu's method maximizes)
    weight1 = np.sum(labels == 0) / len(labels)
    weight2 = np.sum(labels == 1) / len(labels)
    inter_class_variance = weight1 * weight2 * (np.mean(signal) - np.mean(background))**2
    # Calculate entropy of the thresholded image
    hist, _ = np.histogram(labels, bins=2)
    hist_norm = hist / np.sum(hist)
    entropy = scipy.stats.entropy(hist_norm)
    metrics = {
        "inter_class_variance": float(inter_class_variance),
        "entropy": float(entropy)
    }
    return labels, threshold, metrics

def classify_gmm_demux(df, logger):
    from hto.external import gmm_demux

    # init
    key_hash = "hash_id"
    htos = df.columns

    # demultiplex
    df_demux = gmm_demux(df, hash_id=key_hash, params="--random_seed 42")

    # get meta data
    classifications = {}
    thresholds = {}
    metrics = {}
    for hto in htos:
        # classify
        classifications[hto] = df_demux[key_hash].str.contains(hto).astype(int)

        # estimate threshold
        min_signal = df.loc[classifications[hto] == 1, hto].min()
        max_background = df.loc[classifications[hto] == 0, hto].max()
        thresholds[hto] = (min_signal + max_background) / 2
        logger.debug(f"Threshold for {hto}: {thresholds[hto]}")

        # add to metrics
        metrics[hto] = {
            "min_signal": min_signal,
            "max_background": max_background,
        }

    return classifications, thresholds, metrics



import pandas as pd
import anndata as ad
import scanpy as sc
import warnings
warnings.filterwarnings('ignore', module='anndata')

from . import tl
from .normalise import normalise
from .denoise import denoise
from .demux import demux
from .heuristics import heuristics
from ._defaults import DEFAULTS
from ._utils import write_h5ad_safe, write_csv_safe, test_write, subset_whitelist, get_arg, user_input_error_decorator
from ._cluster_background import assert_background
from ._classify import assert_demux
from ._logging import get_logger
from ._exceptions import UserInputError
from .report import report_safe

@user_input_error_decorator
def dnd(
    adata_hto: ad.AnnData,
    adata_hto_raw: ad.AnnData = None,
    adata_gex: ad.AnnData = None,
    adata_background: ad.AnnData = None,
    adata_out: str = None,
    csv_out: str = None,
    path_report: str = None,
    show_report: bool = False,
    _as_cli: bool = False,  # required when run as cli
    **kwargs
):
    f"""Perform normalization and demultiplexing on the provided filtered and raw AnnData objects.

    Args:
        adata_filtered (AnnData): AnnData object with filtered counts
        adata_raw (AnnData): AnnData object with raw counts
        verbose (int, optional): Verbosity level. Default is 1.
    Returns:
        demux_adata (AnnData): An AnnData object containing the results of the demultiplexing.
    """

    # SET PARAMS
    inplace = get_arg("inplace", kwargs, DEFAULTS)
    verbose = get_arg("verbose", kwargs, DEFAULTS)
    denoise_version = get_arg("denoise_version", kwargs, DEFAULTS)
    add_key_hashid = get_arg("add_key_hashid", kwargs, DEFAULTS)
    add_key_doublet = get_arg("add_key_doublet", kwargs, DEFAULTS)
    add_key_normalise = get_arg("add_key_normalise", kwargs, DEFAULTS)
    add_key_denoise = get_arg("add_key_denoise", kwargs, DEFAULTS)
    demux_method = get_arg("demux_method", kwargs, DEFAULTS)
    background_method = get_arg("background_method", kwargs, DEFAULTS)
    background_version = get_arg("background_version", kwargs, DEFAULTS)

    # LOGGER
    logger = get_logger("dnd", level=verbose)

    # READ
    # Note - adata_hto can also be a whitelist of barcodes
    if isinstance(adata_hto_raw, str):
        logger.debug(f"Reading raw hto adata from {adata_hto_raw}")
        adata_hto_raw = ad.read_h5ad(adata_hto_raw)
    if isinstance(adata_hto, str):
        if adata_hto.endswith(".h5ad"):
            logger.debug(f"Reading filtered adata from {adata_hto}")
            adata_hto = ad.read_h5ad(adata_hto)
        elif adata_hto.endswith(".csv") or adata_hto.endswith(".csv.gz"):
            logger.debug(f"Reading whitelist from {adata_hto}")
            whitelist = pd.read_csv(adata_hto, header=None, index_col=0).index.tolist()
            adata_hto = subset_whitelist(adata_hto_raw, whitelist)
        else:
            raise ValueError(f"Unknown file format for adata_hto: {adata_hto}. Must be anndata (.h5ad) or whitelist (.csv|.csv.gz)")
    if isinstance(adata_background, str):
        logger.debug(f"Reading background adata from {adata_background}")
        adata_background = ad.read_h5ad(adata_background)

    # ASSERTIONS
    # - check that output path is writeable (and .h5ad)
    # - check that parameters are valid
    # - check that keys do not exist in adata if run as cli
    test_write(adata_out, filetype="h5ad", create_folder=True, _require_write=_as_cli)
    test_write(csv_out, filetype="csv", create_folder=True, _require_write=_as_cli)
    test_write(path_report, filetype="pdf", create_folder=True, _require_write=_as_cli)
    assert_background(background_method)
    assert_demux(demux_method)
    if _as_cli:
        assert adata_out is not None, "Output path must be provided using parameter --output-path"
        assert add_key_normalise not in adata_hto.layers, f"Key {add_key_normalise} already exists in adata. Add option --add-key-normalise to change the key."
        assert add_key_denoise not in adata_hto.layers, f"Key {add_key_denoise} already exists in adata. Add option --add-key-denoise to change the key."

    # GET DATA
    if adata_gex is not None:
        if isinstance(adata_gex, str) and background_version in ["v1", "v3"]:
            logger.debug(f"Reading gex adata from {adata_gex}")
            if adata_gex.endswith(".h5"):
                adata_gex = sc.read_10x_h5(adata_gex)
            elif adata_gex.endswith(".h5ad"):
                adata_gex = ad.read_h5ad(adata_gex)
            else:
                raise UserInputError(f"Unknown file format for adata_gex: {adata_gex}. Must be anndata (.h5ad) or 10x h5 (.h5)")

    # LOG
    logger.info(f"Starting DND: Normalise (build-background: {background_version}) -> Denoise (background-dection: {background_method} | version: {denoise_version}) -> Demux (method: {demux_method})")

    # RUN
    adata_hto = normalise(
        adata_hto=adata_hto,
        adata_hto_raw=adata_hto_raw,
        adata_gex=adata_gex,
        adata_background=adata_background,
        background_version=background_version,
        inplace=inplace,
        verbose=verbose,
        add_key_normalise=add_key_normalise,
        use_layer=get_arg("use_layer", kwargs, DEFAULTS),
        pseudocount=get_arg("pseudocount", kwargs, DEFAULTS),
    )

    adata_hto = denoise(
        adata_hto=adata_hto,
        inplace=inplace,
        verbose=verbose,
        use_layer=add_key_normalise,
        add_key_denoise=add_key_denoise,
        background_method=background_method,
        denoise_version=denoise_version,
        covariates=get_arg("covariates", kwargs, DEFAULTS),
        design=get_arg("design", kwargs, DEFAULTS),
        kwargs_denoise=get_arg("kwargs_denoise", kwargs, DEFAULTS),
    )

    adata_hto = demux(
        adata_hto=adata_hto,
        inplace=inplace,
        verbose=verbose,
        use_layer=add_key_denoise,
        demux_method=demux_method,
        key_normalise=add_key_normalise,
        enforce_larger_than_background=get_arg("enforce_larger_than_background", kwargs, DEFAULTS),
        add_key_hashid=add_key_hashid,
        add_key_doublet=add_key_doublet,
        add_key_labels=get_arg("add_key_labels", kwargs, DEFAULTS),
    )

    # SAVE
    write_h5ad_safe(adata_hto, adata_out, create_folder=True, _require_write=_as_cli)
    write_csv_safe(adata_hto, csv_out, key_hashid=add_key_hashid, key_doublet=add_key_doublet, create_folder=True)

    # REPORT
    if add_key_normalise is None or add_key_denoise is None:
        logger.warning("Skipping report. Require parameters 'add_key_normalise' (--add-key-normalise) and 'add_key_denoise' (--add-key-denoise) to generate report.")
    elif path_report is not None or show_report:

        # get background
        if adata_background is None:
            adata_background = subset_whitelist(adata_hto_raw, adata_hto.uns["dnd"]["normalise"]["params"]["background"])

        # run report
        report_safe(
            adata_hto=adata_hto,
            adata_background=adata_background,
            adata_hto_raw=adata_hto_raw,
            adata_gex=adata_gex,
            path_report=path_report,
            use_key_normalise=add_key_normalise,
            use_key_denoise=add_key_denoise,
            show=show_report,
            verbose=verbose,
        )

    return adata_hto


import os
from click.testing import CliRunner
import pytest
import numpy as np
import pandas as pd
import anndata as ad

from hto.cli import cli
from hto._defaults import DEFAULTS

@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_cli(mock_hto_data):
    """
    Test if normalisation works.
    """
    # Get mock data
    path_filtered = mock_hto_data['path_filtered']
    path_raw = mock_hto_data['path_raw']
    adata_out = os.path.join(mock_hto_data['path'], "adata.h5ad")
    csv_out = os.path.join(mock_hto_data['path'], "hash_ids.csv")

    # run in cli
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "--adata-hto", path_filtered,
            "--adata-hto-raw", path_raw,
            "--adata-out", adata_out,
            "--csv-out", csv_out,
            "--background-version", "v2",
            "--demux-method", "otsu",
        ]
    )
    assert result.exit_code == 0, f"Exit code is {result.exit_code}: {result.output} | {result.exception}"
    assert os.path.exists(adata_out)
    assert os.path.exists(csv_out)

    # check adata
    adata = ad.read_h5ad(adata_out)
    assert "normalised" in adata.layers.keys()
    assert "denoised" in adata.layers.keys()
    assert DEFAULTS["add_key_hashid"] in adata.obs.keys()
    assert DEFAULTS["add_key_doublet"] in adata.obs.keys()
    assert adata.uns["dnd"]["demux"]["params"]["demux_method"] == "otsu"

    # check csv
    df = pd.read_csv(csv_out, index_col=None)
    assert "hash_id" in df.columns
    assert "doublet_info" in df.columns
    assert "barcode" in df.columns
    assert df.shape[0] == adata.shape[0]

    # check labels
    true_labels = adata.obs['hto_classification']
    predicted_labels = adata.obs['hash_id']
    overall_accuracy = np.mean(predicted_labels == true_labels)
    assert overall_accuracy > 0.8, f"Overall accuracy is only {overall_accuracy:.2f}"


@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_faulty_inputs(mock_hto_data):
    """
    Test if normalisation works.
    """
    # Get mock data
    path_filtered = mock_hto_data['path_filtered']
    path_raw = mock_hto_data['path_raw']
    adata_out = os.path.join(mock_hto_data['path'], "adata.h5ad")

    runner = CliRunner()

    # wrong output path
    result = runner.invoke(
        cli,
        [
            "--adata-hto", path_filtered,
            "--adata-hto-raw", path_raw,
            "--adata-out", "temp",
        ]
    )
    assert result.exit_code == 1, f"Exit code is {result.exit_code}: {result.output}"

    # wrong output path csv
    result = runner.invoke(
        cli,
        [
            "--adata-hto", path_filtered,
            "--adata-hto-raw", path_raw,
            "--adata-out", "temp.h5ad",
            "--csv-out", "temp.false"
        ]
    )
    assert result.exit_code == 1, f"Exit code is {result.exit_code}: {result.output}"

    # wrong method
    result = runner.invoke(
        cli,
        [
            "--adata-hto", path_filtered,
            "--adata-hto-raw", path_raw,
            "--adata-out", adata_out,
            "--demux-method", "wrong_method",
        ]
    )
    assert result.exit_code == 1, f"Exit code is {result.exit_code}: {result.output}"

    # key already exists
    adata = ad.read_h5ad(path_filtered)
    adata.layers["normalised"] = adata.X
    adata.write(path_filtered)
    result = runner.invoke(
        cli,
        [
            "--adata-hto", path_filtered,
            "--adata-hto-raw", path_raw,
            "--adata-out", adata_out,
        ]
    )
    assert result.exit_code == 1, f"Exit code is {result.exit_code}: {result.output}"


import pytest
import numpy as np
import pandas as pd
import anndata as ad
import yaml
from sklearn.datasets import make_blobs

from hto import normalise, denoise, demux
from hto._classify import SUPPORTED_DEMUX_METHODS, classify
from hto._exceptions import AnnDataFormatError


@pytest.fixture
def adata_hto():
    x = np.array([
        [100, 0, 0],
        [0, 100, 0],
        [0, 0, 100],
        [100, 100, 0],
        [0, 100, 100],
        [0, 0, 0],
        [0, 0, 0],
    ]) + 0.1 # (must be float)
    labels = np.array([
        [1, 0, 0],
        [0, 1, 0],
        [0, 0, 1],
        [1, 1, 0],
        [0, 1, 1],
        [0, 0, 0],
        [0, 0, 0],
    ])
    var_names = [f"hto_{i}" for i in range(x.shape[1])]
    obs_names = [f"cell_{i}" for i in range(x.shape[0])]
    true_labels = ["hto_0", "hto_1", "hto_2", "doublet", "doublet", "negative", "negative"]
    true_doublet_info = ["", "", "", "hto_0,hto_1", "hto_1,hto_2", "", ""]
    adata = ad.AnnData(X=x, obs=pd.DataFrame({
        "true_labels": true_labels,
        "true_doublet_info": true_doublet_info,
    }, index=obs_names), var=pd.DataFrame(index=var_names))
    adata.layers["labels"] = labels
    return adata


def test_demux(adata_hto):
    """
    Test the demux function for demultiplexing using different methods.
    """

    # All should work
    for method in SUPPORTED_DEMUX_METHODS:

        # Run demux
        adata_demux = demux(
            adata_hto=adata_hto,
            demux_method=method,
            enforce_larger_than_background=False,  # <- no normalised layer available
            add_key_labels="demux_labels",
            use_layer=None,
            inplace=False,
        )

        # Check results
        assert "hash_id" in adata_demux.obs.columns
        assert "doublet_info" in adata_demux.obs.columns
        assert "demux_labels" in adata_demux.layers
        assert "metrics" in adata_demux.uns["dnd"]["demux"]
        assert "thresholds" in adata_demux.uns["dnd"]["demux"]
        thresholds = adata_demux.uns["dnd"]["demux"]["thresholds"]
        assert len(thresholds) == adata_hto.X.shape[1]
        assert isinstance(thresholds["hto_1"], float)
        # check classification
        assert all(adata_demux.obs["hash_id"] == adata_demux.obs["true_labels"])
        assert np.all(adata_demux.layers["demux_labels"] == adata_hto.layers["labels"])
        # check doublet_info
        assert np.all(adata_demux.obs["doublet_info"] == adata_demux.obs["true_doublet_info"])


@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
@pytest.mark.parametrize("demux_method", ["kmeans", "gmm", "otsu", "gmm_demux"])
def test_classify(mock_hto_data, demux_method):
    """
    Test the clustering and evaluation of HTO data.
    Checks the following:
    1. The number of unique labels is exactly 2.
    2. The identified positive cluster is either 0 or 1.
    3. For "kmeans":
        - The metrics dictionary contains 'silhouette_score' and 'davies_bouldin_index'.
        - The silhouette score is positive
        - The Davies-Bouldin index is positive.
    4. For the "gmm" method:
        - The metrics dictionary contains 'bic' and 'log_likelihood'.
        - The log-likelihood is negative, as expected.
    5. For the "otsu" method:
        - The metrics dictionary contains 'threshold', 'inter_class_variance', and 'entropy'.
        - The threshold is a float greater than 0.
    Parameters:
         mock_hto_data: An AnnData object containing the denoised data.
         method: A string indicating the clustering method to use ("kmeans" or "gmm").
    """

    adata_filtered = mock_hto_data['filtered']
    adata_hto_raw = mock_hto_data['raw']
    adata_normalised = normalise(
        adata_filtered,
        adata_hto_raw=adata_hto_raw,
        background_version="v2",
        add_key_normalise="normalised"
    )
    adata_denoised = denoise(
        adata_normalised,
        denoise_version="v1",
        add_key_denoise="denoised",
        use_layer="normalised"
    )
    df = adata_denoised.to_df("denoised")

    classifications, thresholds, metrics = classify(
        data=df,
        demux_method=demux_method,
    )

    for hto, labels in classifications.items():
        # Check if the labels are binary
        assert len(np.unique(labels)) == 2

    for hto, thresh in thresholds.items():
        # Check if the threshold is a float
        assert isinstance(thresh, float)
        if demux_method == "gmm_demux":
            print("DEBUG")
            print(pd.DataFrame(classifications))
        assert thresh > 0

    # metric specific
    for hto, _metrics in metrics.items():
        if demux_method == "kmeans":
            assert 'silhouette_score' in _metrics
            assert 'davies_bouldin_index' in _metrics
            assert _metrics['silhouette_score'] > 0
            assert _metrics['davies_bouldin_index'] > 0
        elif demux_method == "gmm":
            assert 'bic' in _metrics
            assert 'log_likelihood' in _metrics
            assert _metrics['log_likelihood'] < 0
        elif demux_method == "otsu":
            assert 'inter_class_variance' in _metrics
            assert 'entropy' in _metrics
        elif demux_method == "gmm_demux":
            assert 'min_signal' in _metrics
            assert 'max_background' in _metrics


def test_enforce_larger_than_background(adata_hto):
    """
    Test if the enforce_larger_than_background option works.
    """
    # Replace values with negative values to test
    adata_neg = adata_hto.copy()
    adata_neg.X = adata_neg.X - 1000
    adata_neg.layers["normalised"] = adata_neg.X.copy()

    # Enforce
    adata_demux_true = demux(
        adata_hto=adata_neg,
        demux_method="otsu",
        enforce_larger_than_background=True,
        key_normalise="normalised",
        use_layer=None,
        inplace=False,
    )
    assert np.all(adata_demux_true.obs["hash_id"] == "negative")

    # Do not enforce
    adata_demux_false = demux(
        adata_hto=adata_neg,
        demux_method="otsu",
        enforce_larger_than_background=False,
        key_normalise="normalised",
        use_layer=None,
        inplace=False,
    )
    assert np.all(adata_demux_false.obs["hash_id"] == adata_hto.obs["true_labels"])


@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_faulty_data(mock_hto_data):
    """
    Test if normalisation works.
    """
    # Get mock data
    adata_filtered = mock_hto_data['filtered']
    adata_raw = mock_hto_data['raw']

    # Skip preprocessing
    with pytest.raises(AnnDataFormatError):
        denoise(adata_filtered)


import pytest
import numpy as np
import pandas as pd
import anndata as ad
from hto._exceptions import UserInputError

import hto

@pytest.fixture
def hto_and_gex():
    # generate some test data
    # cell 1: real cell
    # cell 2: background  barcode
    # cell 3: discard
    n = 50
    cell_ids = np.array([f"cell_{i}" for i in range(2 * n)])
    x_hto = np.ones((2 * n, 3))
    x_gex = np.concatenate([
        # non-empty
        np.random.randint(100, 200, (n, 3)),
        # empty
        np.random.randint(0, 10, (n, 3))
    ])

    obs = pd.DataFrame(index=cell_ids)
    adata_hto = ad.AnnData(X=x_hto, obs=obs)
    adata_gex = ad.AnnData(X=x_gex, obs=obs)

    return adata_hto, adata_gex

@pytest.fixture
def hto_and_raw():
    cell_ids = np.array([f"cell_{i}" for i in range(1, 7)])
    x_hto = np.ones((3, 3)) + 1
    x_hto_raw = np.ones((3, 3))

    obs = pd.DataFrame(index=cell_ids)
    adata_hto = ad.AnnData(X=x_hto, obs=obs.iloc[:3])
    adata_hto_raw = ad.AnnData(X=np.concatenate([x_hto, x_hto_raw]), obs=obs)

    return adata_hto, adata_hto_raw


def test_build_background_v1(hto_and_gex):
    n = 50

    adata_hto, adata_gex = hto_and_gex
    adata_background = hto.tl.build_background(
        "v1",
        adata_hto_raw=adata_hto,
        adata_gex=adata_gex,
        min_umi=300,
        _run_assert=False
    )

    assert adata_hto.shape[0] == 2 * n
    assert adata_gex.shape[0] == 2 * n
    assert adata_background.shape[0] == n
    # First 'n' cells are background
    set_background = set(adata_background.obs_names)
    set_expected = set([f'cell_{i}' for i in range(n)])
    assert set_background == set_expected


def test_build_background_v2(hto_and_raw):
    adata_hto, adata_hto_raw = hto_and_raw

    adata_background = hto.tl.build_background(
        "v2",
        adata_hto=adata_hto,
        adata_hto_raw=adata_hto_raw,
        _run_assert=False
    )

    assert adata_hto.shape[0] == 3
    assert adata_hto_raw.shape[0] == 6
    assert adata_background.shape[0] == 3
    # check that sets are identicel
    set_background = set(adata_background.obs_names)
    set_expected = set(['cell_1', 'cell_2', 'cell_3'])
    assert set_background == set_expected

    # check faulty data
    with pytest.raises(UserInputError):
        adata_background = hto.tl.build_background(
            "v2",
            adata_hto=adata_hto,
            adata_hto_raw=adata_hto,
            _run_assert=False
        )

def test_build_background_v3(hto_and_gex):
    adata_hto_raw, adata_gex = hto_and_gex
    # - 0-30 are real cells
    # - 30-50 are background
    # - 50+ are empty
    adata_hto = adata_hto_raw.copy()[:30]

    adata_background = hto.tl.build_background(
        "v3",
        adata_hto=adata_hto,
        adata_hto_raw=adata_hto_raw,
        adata_gex=adata_gex,
        k_gex_cells=10,
        _run_assert=False
    )

    assert adata_hto.shape[0] == 30
    assert adata_hto_raw.shape[0] == 100
    assert adata_gex.shape[0] == 100
    assert adata_background.shape[0] == 40



#!/usr/bin/env python
# coding: utf-8

import pandas as pd
from pprint import pformat
import scipy.sparse
import anndata as ad

from ._classify import classify, assert_demux
from ._defaults import DEFAULTS, DESCRIPTIONS
from ._logging import get_logger
from ._meta import add_meta
from ._utils import get_layer
from ._exceptions import UserInputError

def demux(
    adata_hto: ad.AnnData,
    demux_method: str = DEFAULTS["demux_method"],
    use_layer: str = DEFAULTS["use_layer"],
    key_normalise: str = DEFAULTS["add_key_normalise"],
    enforce_larger_than_background: bool = DEFAULTS["enforce_larger_than_background"],
    add_key_hashid: str = DEFAULTS["add_key_hashid"],
    add_key_doublet: str = DEFAULTS["add_key_doublet"],
    add_key_labels: str = DEFAULTS["add_key_labels"],
    inplace: bool = DEFAULTS["inplace"],
    verbose: int = DEFAULTS["verbose"],
):
    f"""Classify HTOs as singlets (assign to HTO), doublets, or negatives.

    Use a 2-component K-means, GMM, or Otsu threshold method to categorize cells based on their HTO classifications.

    Args:
        adata_hto (ad.AnnData): {DESCRIPTIONS["adata_hto"]}
        demux_method (str): {DESCRIPTIONS["demux_method"]}
        use_layer (str): {DESCRIPTIONS["use_layer"]}
        add_key_hashid (str): {DESCRIPTIONS["add_key_hashid"]}
        add_key_doublet (str): {DESCRIPTIONS["add_key_doublet"]}
        inplace (bool): {DESCRIPTIONS["inplace"]}
        verbose (int): {DESCRIPTIONS["verbose"]}

    Returns:
        AnnData: An AnnData object containing the results of the demultiplexing in .obs.
    """

    # debug - print parameters
    logger = get_logger("demux", level=verbose)
    params = {k: v.shape if isinstance(v, ad.AnnData) else v for k, v in locals().items()}
    params_str = pformat(params, indent=4)
    logger.debug(f"Parameters:\n{params_str}")

    # setup data
    adata_hto, _ = get_layer(
        adata=adata_hto,
        use_layer=use_layer,
        numpy=True,
        float=True,
        inplace=inplace,
    )

    # assertions
    assert_demux(demux_method)
    if add_key_hashid in adata_hto.obs.columns:
        logger.warning(f"Column '{add_key_hashid}' already exists in adata.obs. Overwriting and storing previous columns under '{add_key_hashid}_archive'.")
        adata_hto.obs[f"{add_key_hashid}_archive"] = adata_hto.obs[add_key_hashid]
        adata_hto.obs.drop(add_key_hashid, axis=1, inplace=True)
    
    if enforce_larger_than_background:
        if key_normalise not in adata_hto.layers.keys():
            raise UserInputError(
                f"Layer '{key_normalise}' not found in adata. Please normalise the data first or set 'enforce_larger_than_background' to False."
            )

    # get data
    df_umi = adata_hto.to_df(layer=use_layer)
    assert all([t == float for t in df_umi.dtypes]), "Denoised data must be float."

    # Get classifications for each HTO
    logger.debug(f"Starting demultiplexing using '{demux_method}'...")
    classifications, thresholds, metrics = classify(
        data=df_umi,
        demux_method=demux_method,
        verbose=verbose,
    )

    # Enforce that positively labeled cells are larger than background. This can be skewed due to the technical noise correction of small cells.
    if enforce_larger_than_background:
        logger.debug("Enforcing larger than background...")
        df_normalised = adata_hto.to_df(layer=key_normalise)
        for hto in df_normalised.columns:
            classifications[hto][df_normalised[hto] <= 0] = 0

    # Init results
    logger.debug("Assigning labels...")
    labels_df = pd.DataFrame(classifications, index=df_umi.index)
    result_df = pd.DataFrame("", index=labels_df.index, columns=[add_key_hashid, add_key_doublet])

    # Get cell labels
    result_df.loc[:, add_key_hashid] = labels_df.idxmax(axis=1)
    result_df.loc[labels_df.sum(axis=1) == 0, add_key_hashid] = "negative"
    result_df.loc[labels_df.sum(axis=1) >= 2, add_key_hashid] = "doublet"

    # Get doublet info
    logger.debug("Assigning doublet info...")
    result_df.loc[:, add_key_doublet] = labels_df.apply(
        lambda row: ",".join([r for r in row[row == 1].index if row.sum() >= 2]), axis=1
    )

    # Append to AnnData
    if add_key_labels is not None:
        adata_hto.layers[add_key_labels] = scipy.sparse.csr_matrix(labels_df.values)
    adata_hto.obs = pd.concat([adata_hto.obs, result_df], axis=1)

    # add metadata
    adata_hto = add_meta(
        adata_hto,
        step="demux",
        params=dict(
            demux_method=demux_method,
        ),
        metrics=metrics,
        thresholds=thresholds,
    )

    pct_doublet = (result_df[add_key_hashid] == "doublet").mean() * 100
    pct_negative = (result_df[add_key_hashid] == "negative").mean() * 100
    logger.info(f"Demultiplexing completed ({pct_doublet:.1f}% Doublets, {pct_negative:.1f}% Negatives). Hash IDs stored in '{add_key_hashid}'.")
    return adata_hto



import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.svm import LinearSVR

from ._defaults import DEFAULTS, DESCRIPTIONS
from ._utils import get_arg
from ._logging import get_logger

SUPPORTED_DENOISE_VERSIONS = ["v1", "v2"]

def remove_batch_effect(
    denoise_version: str,
    verbose: int = DEFAULTS["verbose"],
    **kwargs,
):
    f"""
    Remove batch effect of adt count data (x) using a cell-by-cell noise vector (covariates).

    - v1: Linear regression of x on covariates and design matrix.
    - v2: Support Vector Regression (SVR) of x on covariates.
    """
    logger = get_logger("remove_batch_effect", level=verbose)

    if denoise_version == "v1":
        logger.warning(f"Not recommended. Using denoise version: {denoise_version}. Plot 'hto_dnd.pl.technical_noise(adata)' to verify that the denoising is working as expected.")
        return remove_batch_effect_v1(
            x=kwargs["x"],
            covariates=kwargs["covariates"],
            design=get_arg("design", kwargs, DEFAULTS),
        )
    elif denoise_version == "v2":
        return remove_batch_effect_v2(
            x=kwargs["x"],
            covariates=kwargs["covariates"],
            kwargs_denoise=kwargs["kwargs_denoise"],
        )
    else:
        raise ValueError(f"Invalid version: {denoise_version}. Must be 'v1'.")


def remove_batch_effect_v1(
    x: np.ndarray,
    covariates: np.ndarray,
    design: np.ndarray,
) -> np.ndarray:

    # assertions
    assert isinstance(x, np.ndarray), "Input matrix must be a NumPy array"
    if design is None:
        design = np.ones((x.shape[0], 1))
    else:
        design = np.asarray(design)

    # Process covariates (in our case, this is the background means)
    covariates = np.asarray(covariates).reshape(-1, 1)

    # Combine design and covariates
    X_combined = np.column_stack([design, covariates])

    # Fit linear model
    model = LinearRegression(fit_intercept=False)
    model.fit(X_combined, x)

    # Extract coefficients related to batch effects
    beta = model.coef_[:, design.shape[1] :]
    # beta = model.coef_

    # Broadcast the multiplication. here beta is the coefficient of the regression and covariates is the baclground means. their multiplication is just the prediction of how much technical noise there is and then after we predict that, we subtract it from x (the normalized matrix) to get the corrected matrix
    correction = covariates @ beta.T

    # Subtract the correction from x to remove the batch effect
    x_corrected = x - correction

    # Store metadata
    meta = {
        "coefs": model.coef_,
    }

    return x_corrected, meta


def remove_batch_effect_v2(
    x,
    covariates,
    kwargs_denoise,
) -> np.ndarray:

    # assertions
    assert isinstance(x, np.ndarray), "Input matrix must be a NumPy array"
    assert isinstance(covariates, np.ndarray), "Covariates must be a NumPy array"
    covariates = np.asarray(covariates).reshape(-1, 1)

    coefs = []
    x_corrected = np.zeros_like(x)
    for i in range(x.shape[1]):
        x_i = x[:, i]

        # Fit SVR model
        model = LinearSVR(fit_intercept=True, **kwargs_denoise)
        model.fit(covariates, x_i)
        coefs.append([float(model.intercept_), float(model.coef_[0])])

        # apply correction
        pred = model.predict(covariates)
        x_corrected[:, i] = x_i - pred

    # Store metadata
    meta = {
        "coefs": coefs,
    }

    return x_corrected, meta


import pytest
from hto.external import gmm_demux

@pytest.mark.parametrize("mock_hto_data", [{'n_cells': 100}], indirect=True)
def test_gmm_demux(mock_hto_data):
    adata_filtered = mock_hto_data['filtered']
    df_demux = gmm_demux(adata_filtered.to_df(), hash_id="hash_id_gmm_demux")
    assert "hash_id_gmm_demux" in df_demux.columns

    adata_filtered.obs = adata_filtered.obs.join(df_demux, how="left")
    assert not any(adata_filtered.obs["hash_id_gmm_demux"].isna())

    accuracy = (adata_filtered.obs["hash_id_gmm_demux"] == adata_filtered.obs["hto_classification"]).mean()
    assert accuracy > .6, f"Classification accuracy appears random, got {accuracy:.2f} instead of > 0.6"
